// diskerase.cc
//
// g++ -o diskerase.exe -g -O2 -Wall -Wuninitialized diskerase.cc
//
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>
#include <assert.h>

uint64_t fsize(int fd)
{
  off64_t current = lseek64(fd,0,SEEK_CUR);
  off64_t end     = lseek64(fd,0,SEEK_END);
  lseek64(fd,current,SEEK_SET);
  return end;
}

int main(int argc,char*argv[])
{
  if ((argc < 2)||(argc > 4))
    {
      fprintf(stderr,"Usage: %s /dev/hdXXX [offsetMiB] [sizeMiB]\n",argv[0]);
      exit(1);
    }

  setbuf(stdout,0);
  setbuf(stderr,0);

  int tty = isatty(1);

  char*diskname = argv[1];
  int  offsetMiB = 0;
  int  sizeMiB = 0;

  if (argc >= 3)
    offsetMiB = atoi(argv[2]);

  if (argc >= 4)
    sizeMiB = atoi(argv[3]);

  int rfd = -1;

  rfd = open(diskname,O_RDWR|O_LARGEFILE);

  if (rfd < 0)
    {
      fprintf(stderr,"Cannot open file \'%s\', errno: %d (%s)\n",diskname,errno,strerror(errno));
      return -1;
    }

  off64_t mib = 1024*1024;
  off64_t gib = 1024*1024*1024;

  uint64_t size = fsize(rfd);

  if (sizeMiB != 0)
    size = sizeMiB*mib;

  printf("File size: %lld\n",(long long int)size);

#if 0
  int err = posix_fadvise(rfd,0,0,POSIX_FADV_SEQUENTIAL|POSIX_FADV_NOREUSE);
  if (err != 0)
    {
      fprintf(stderr,"Cannot fadvise(), errno: %d (%s)\n",errno,strerror(errno));
      return -1;
    }
#endif

  int countSlow = 0;
  int countUnwriteable = 0;
  int countShort = 0;
  double longestWriteTime = 0;
  double maxRate = -HUGE_VAL;
  double minRate = +HUGE_VAL;

  const double kExtremelySlowWriteTime = 200.0; // sec

  int incr = 10;
  int position = offsetMiB;

  uint64_t countBytes = offsetMiB*mib;
  uint64_t totalCountBytes = 0;

  time_t timeBegin;
  time(&timeBegin);

  int eof = 0;
  while (!eof)
    {
      off64_t seek = position * mib;

      off64_t current = lseek64(rfd,0,SEEK_CUR);
      if (current < 0)
        {
           fprintf(stderr,"At %d MiBytes: Cannot determine position in file \'%s\' to %d MiBytes (%lld), errno: %d (%s)\n",position,diskname,position,(long long int)seek,errno,strerror(errno));
          break;
        }

      if (current != seek)
        {
          off64_t status = lseek64(rfd,seek,SEEK_SET);
          if (status < 0)
            {
              fprintf(stderr,"At %d MiBytes: Cannot seek file \'%s\' to %d MiBytes (%lld), errno: %d (%s)\n",position,diskname,position,(long long int)seek,errno,strerror(errno));

              break;
            }
        }
          
      double countbytes = 0;
      double counttime  = 0;

      double maxwritetime = 0;

      int shouldWrite = incr*mib;
      int didWrite = 0;

      while (shouldWrite > 0)
        {
          struct timeval tstart, tend;
          int bufsize = 1024*1024;
          char buf[bufsize];
          bool maybeRewrite = false;

          static char xxx = 1;
          memset(buf, xxx++, bufsize);

          errno = 0;

          int towrite = shouldWrite;

          if (towrite > bufsize)
            towrite = bufsize;

          gettimeofday(&tstart,0);
          
          int rd = write(rfd,buf,towrite);

          int xx = fdatasync(rfd);
          if (xx != 0)
             rd = xx;
          
          gettimeofday(&tend,0);
          
          long diffs = tend.tv_sec  - tstart.tv_sec;
          long diffu = tend.tv_usec - tstart.tv_usec;
          double writetime = diffs + 0.000001 * diffu;

          if (writetime > longestWriteTime)
            longestWriteTime = writetime;

          if (writetime > maxwritetime)
            maxwritetime = writetime;

          if (writetime > 1.0)
            {
              fprintf(stderr,"At %d MiBytes: Slow write around offset %.0f, write time: %.3f sec\n",position,countbytes,writetime);
              countSlow++;

              maybeRewrite = true;
            }
          
          if (rd == 0)
            {
               fprintf(stderr,"At %d MiBytes: EOF, remaining: %lld.\n",position,(long long int)(size-countBytes));
              eof = 1;
              break;
            }
          else if (rd < 0)
            {
              if (errno != EIO)
                {
                  fprintf(stderr,"At %d MiBytes: Unwriteable area at offset %d: Unexpected write error, errno: %d (%s), giving up.\n",position,didWrite,errno,strerror(errno));
                  countUnwriteable++;
                  eof = 1;
                  break;
                }

              fprintf(stderr,"At %d MiBytes: Unwriteable area at offset %d: I/O error, errno: %d (%s)\n",position,didWrite,errno,strerror(errno));
              countUnwriteable++;

              if (writetime > kExtremelySlowWriteTime)
                {
                  fprintf(stderr,"At %d MiBytes: Extremely slow write at offset %d: %.3f sec, giving up.\n",position,didWrite,writetime);
                  countSlow++;
                  eof = 1;
                }

              break;
            }

          shouldWrite -= rd;
          didWrite    += rd;

          countBytes += rd;
          totalCountBytes += rd;

          if (rd != towrite)
            {
              if (size - countBytes > 0)
                {
                   fprintf(stderr,"At %d MiBytes: Short write at offset %d, remaining %lld...\n",position,didWrite,(long long int)(size-countBytes));
                  countShort++;
                }
              else
                {
                  fprintf(stderr,"At %d MiBytes: End of file: short write at offset %d.\n",position,didWrite);
                  eof = true;
                }

              maybeRewrite = false;
            }

          if (writetime > kExtremelySlowWriteTime)
            {
              fprintf(stderr,"At %d MiBytes: Extremely slow write at offset %d: %.3f sec, giving up.\n",position,didWrite,writetime);
              countSlow++;
              eof = 1;
              break;
            }

          countbytes += rd;
          counttime  += writetime;

#if 0
          printf("%d %f\n",rd,writetime);
          if (i>100)
            {
              stop = 1;
              break;
            }
#endif
        }

      double rate = countbytes/counttime;

      if (rate > maxRate)
        maxRate = rate;

      if (rate < minRate)
        minRate = rate;

      if (tty)
        printf("At %d MiBytes: %.0f bytes, max write times: %.3f and %.3f sec, data rate: %.0f MiBytes/sec\r",position,countbytes,maxwritetime,longestWriteTime,rate/(1024*1024));

      position += incr;

      if (sizeMiB > 0)
        if (position > sizeMiB)
          break;
    }

  time_t timeEnd;
  time(&timeEnd);

  fprintf(stderr,"Summary: disk size %.0f GiBytes, %d slow writes, %d short writes, %d unwriteable areas, data rate: min: %.0f, max: %.0f, average: %.0f MiBytes/sec, longest write time %.3f sec\n",size/(double)gib,countSlow,countShort,countUnwriteable,minRate/mib,maxRate/mib,(totalCountBytes/(double)mib/(double)(timeEnd-timeBegin)),longestWriteTime);

  return 0;
}

// end file
