// diskscan.cc
//
// g++ -o diskscrub.exe -g -O2 -Wall -Wuninitialized diskscrub.cc
//
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>
#include <assert.h>

#include <sys/syscall.h>

#include <string>
#include <vector>

#ifdef SYS_ioprio_set
#define HAVE_IOPRIO 1
#endif

#ifdef HAVE_IOPRIO

/*
 * Gives us 8 prio classes with 13-bits of data for each class
 */
#define IOPRIO_BITS             (16)
#define IOPRIO_CLASS_SHIFT      (13)
#define IOPRIO_PRIO_MASK        ((1UL << IOPRIO_CLASS_SHIFT) - 1)

#define IOPRIO_PRIO_CLASS(mask) ((mask) >> IOPRIO_CLASS_SHIFT)
#define IOPRIO_PRIO_DATA(mask)  ((mask) & IOPRIO_PRIO_MASK)
#define IOPRIO_PRIO_VALUE(class, data)  (((class) << IOPRIO_CLASS_SHIFT) | data)

#define ioprio_valid(mask)      (IOPRIO_PRIO_CLASS((mask)) != IOPRIO_CLASS_NONE)

/*
 * These are the io priority groups as implemented by CFQ. RT is the realtime
 * class, it always gets premium service. BE is the best-effort scheduling
 * class, the default for any process. IDLE is the idle scheduling class, it
 * is only served when no one else is using the disk.
 */
enum {
        IOPRIO_CLASS_NONE,
        IOPRIO_CLASS_RT,
        IOPRIO_CLASS_BE,
        IOPRIO_CLASS_IDLE,
};

/*
 * 8 best effort priority levels are supported
 */
#define IOPRIO_BE_NR    (8)

enum {
        IOPRIO_WHO_PROCESS = 1,
        IOPRIO_WHO_PGRP,
        IOPRIO_WHO_USER,
};

/*
 * if process has set io priority explicitly, use that. if not, convert
 * the cpu scheduler nice value to an io priority
 */
#define IOPRIO_NORM     (4)

int ioprio_set(int which, int who, int ioprio)
{
   return syscall(SYS_ioprio_set, which, who, ioprio);
}

////// end

#else

#warning "This system does not seem to have the SYS_ioprio_set syscall for changing I/O priorities. Diskscrub will be unable to lower it's priority to improve system responsiveness. Sorry."

#endif

uint64_t fsize(int fd)
{
  off64_t current = lseek64(fd,0,SEEK_CUR);
  off64_t end     = lseek64(fd,0,SEEK_END);
  lseek64(fd, current, SEEK_SET);
  return end;
}

int open_disk(const char* diskname, FILE* flog)
{
   int rfd = open(diskname, O_RDONLY|O_LARGEFILE);

   if (rfd < 0)
      {
         fprintf(flog, "Cannot open file \'%s\', errno: %d (%s)\n", diskname, errno, strerror(errno));
         return -1;
      }

   return rfd;
}

static const uint64_t mib = 1024*1024;
static const uint64_t gib = 1024*1024*1024;

double dtime()
{
   struct timeval t;
   gettimeofday(&t,0);
   return t.tv_sec + 0.000001 * t.tv_usec;
}

void dsleep(double t)
{
   int tmillisec = (int)(t*1000);
   struct timeval tv;
   tv.tv_sec = tmillisec/1000;
   tv.tv_usec = 1000*(tmillisec%1000);
   //printf("sleep %.3f, %d, tv: %d %d\n", t, tmillisec, tv.tv_sec, tv.tv_usec);
   //double t0 = dtime();
   select(0, NULL, NULL, NULL, &tv);
   //double t1 = dtime();
   //double diff = t1-t0;
   //printf("status %d, slept for %.3f sec\n", status, diff);
}

void xctime(const time_t *t, char* buf, int bufsize)
{
   char* s = ctime(t);
   strcpy(buf, s);
   assert(strlen(buf) < (size_t)bufsize);
   char* p = strchr(buf, '\n');
   if (p)
      *p = 0;
   assert(strlen(buf) < (size_t)bufsize);
}

std::string xmlEncodeString(const char* tag, const char* s)
{
   return std::string("")+ "<"+tag+">"+s+"</"+tag+">";
}

std::string xmlEncodeString(const char* tag, const std::string& s)
{
   return std::string("")+ "<"+tag+">"+s+"</"+tag+">";
}

std::string xmlEncodeInt(const char* tag, int value)
{
   char buf[1024];
   sprintf(buf, "<%s>%d</%s>", tag, value, tag);
   assert(strlen(buf)<sizeof(buf));
   return buf;
}

std::string xmlEncodeUint64(const char* tag, uint64_t value)
{
   char buf[1024];
   sprintf(buf, "<%s>%llu</%s>", tag, value, tag);
   assert(strlen(buf)<sizeof(buf));
   return buf;
}

std::string xmlEncodeDouble(const char* tag, double value)
{
   char buf[1024];
   sprintf(buf, "<%s>%g</%s>", tag, value, tag);
   assert(strlen(buf)<sizeof(buf));
   return buf;
}

static std::string gEmailAddress = "";
static bool gDoEmailAlways  = false;
static bool gDoEmailUnhappy = true;
static bool gDoEmailError   = true;

static int gTooManyIOerrors = 10;
static double gSlowReadTime = 1.0; // sec
static double gExtremelySlowReadTime = 200.0; // sec

int scrub_disk(FILE* flog, FILE* fsummary, FILE* fxml, bool verbose, const char* diskname, uint64_t arg_offset, uint64_t arg_size, uint64_t arg_blocksize, time_t arg_time)
{
   char tmp[256];

   int rfd = open_disk(diskname, flog);

   if (rfd < 0)
      return -1;

   uint64_t size = arg_size;
   
   uint64_t sizeMiB = (size+mib-1)/mib;
   uint64_t sizeGiB = (size+gib-1)/gib;
   
   double rate = (arg_size - arg_offset)/(double)arg_time;

   char nowText[1024];
   bool haveFadvise64 = true;

   int countAborted = 0;
   int countSlow = 0;
   int countUnreadable = 0;

   double longestReadTime = 0;
   double maxRate = -HUGE_VAL;
   double minRate = +HUGE_VAL;

   int countSequencialIOerrors = 0;

   unsigned int incr = arg_blocksize / mib;
   unsigned int position = arg_offset / mib;

   uint64_t countBytes = arg_offset;
   uint64_t totalCountBytes = 0;

   double xcountbytes = 0;
   double xcounttime = 0;

   std::string slog;

   time_t timeBegin;
   time(&timeBegin);

   xctime(&timeBegin, nowText, sizeof(nowText));

   fprintf(flog, "%s: Size of \'%s\' is %llu (%llu MiBytes) (%llu GiBytes), allocated time %d sec, data rate %.1f MiBytes/sec\n", nowText, diskname, size, sizeMiB, sizeGiB, (int)arg_time, rate/mib);

   bool eof = false;
   while (!eof)
      {
         off64_t seek = position * mib;

         off64_t current = lseek64(rfd,0,SEEK_CUR);
         if (current < 0)
            {
               fprintf(flog, "At %d MiBytes: Cannot determine position in file \'%s\' to %d MiBytes (%lld), errno: %d (%s)\n",position,diskname,position,seek,errno,strerror(errno));
               break;
            }

         if (current != seek)
            {
               off64_t status = lseek64(rfd,seek,SEEK_SET);
               if (status < 0)
                  {
                     fprintf(flog, "At %d MiBytes: Cannot seek file \'%s\' to %d MiBytes (%lld), errno: %d (%s)\n",position,diskname,position,seek,errno,strerror(errno));

                     break;
                  }

               fprintf(flog,"At %d MiBytes: Reset position from %lld to %lld\n", position, current, seek);
            }
          
         double countbytes = 0;
         double counttime  = 0;

         double maxreadtime = 0;

         int shouldRead = incr*mib;
         int didRead = 0;

         double tstart = dtime();

         if (haveFadvise64)
            {
               errno = 0;
               // start read-ahead of our data
               int err = posix_fadvise64(rfd, seek, shouldRead, POSIX_FADV_WILLNEED);
               // disable read-ahead of our data
               //int err = posix_fadvise64(rfd, seek, shouldRead, POSIX_FADV_RANDOM);
               if ((errno == ENOSYS) || (errno == ENOTTY) || (err != 0 && errno == 0))
                  {
                     fprintf(flog, "fadvise() not available, too bad! (status %d, errno %d (%s))\n", err, errno, strerror(errno));
                     haveFadvise64 = false;
                  }
               else if (err != 0)
                  {
                     fprintf(flog, "Cannot fadvise(), returned %d, errno: %d (%s)\n",err,errno,strerror(errno));
                     return -1;
                  }
            }

         while (shouldRead > 0)
            {
               struct timeval tstart, tend;
               int bufsize = 256*1024;
               char buf[bufsize];

               errno = 0;

               int toread = shouldRead;

               if (toread > bufsize)
                  toread = bufsize;

               gettimeofday(&tstart,0);
          
               int rd = read(rfd,buf,toread);
          
               gettimeofday(&tend,0);
          
               long diffs = tend.tv_sec  - tstart.tv_sec;
               long diffu = tend.tv_usec - tstart.tv_usec;
               double readtime = diffs + 0.000001 * diffu;

               if (readtime > longestReadTime)
                  longestReadTime = readtime;

               if (readtime > maxreadtime)
                  maxreadtime = readtime;

               if (readtime > gSlowReadTime)
                  {
                     fprintf(flog, "At %d MiBytes: Slow read around offset %.0f, read time: %.3f sec\n",position,countbytes,readtime);
                     countSlow++;
                  }
          
               if (rd == 0)
                  {
                     fprintf(flog, "At %d MiBytes: EOF, remaining: %lld.\n",position,size-countBytes);
                     eof = true;
                     break;
                  }
               else if (rd < 0)
                  {
		    countSequencialIOerrors++;

                     if (errno != EIO)
                        {
                           sprintf(tmp, "At %d MiBytes: Unreadable area at offset %d: Unexpected read error, errno: %d (%s), giving up.\n",position,didRead,errno,strerror(errno));
                           fprintf(flog, "%s", tmp);
                           if (fsummary)
                              fprintf(fsummary, "%s", tmp);
                           slog += tmp;

                           countUnreadable++;
                           countAborted++;
                           eof = true;
                           break;
                        }

                     sprintf(tmp, "At %d MiBytes: Unreadable area at offset %d: I/O error %d, errno: %d (%s)\n", position, didRead, countSequencialIOerrors, errno, strerror(errno));
                     fprintf(flog,"%s", tmp);
                     if (fsummary)
                        fprintf(fsummary, "%s", tmp);
                     slog += tmp;

                     countUnreadable++;

                     if (readtime > gExtremelySlowReadTime)
                        {
                           sprintf(tmp, "At %d MiBytes: Extremely slow read at offset %d: %.3f sec, giving up.\n",position,didRead,readtime);
                           fprintf(flog, "%s", tmp);
                           if (fsummary)
                              fprintf(fsummary, "%s", tmp);
                           slog += tmp;

                           countSlow++;
                           countAborted++;
                           eof = true;
                        }

                     if (countSequencialIOerrors > gTooManyIOerrors)
                        {
                           sprintf(tmp, "At %d MiBytes: Too many I/O errors: %d, giving up.\n", position, countSequencialIOerrors);
                           fprintf(flog, "%s", tmp);
                           if (fsummary)
                              fprintf(fsummary, "%s", tmp);
                           slog += tmp;

                           countAborted++;
                           eof = true;
                        }

                     break;
                  }

	       countSequencialIOerrors = 0;

               shouldRead -= rd;
               didRead    += rd;

               countBytes += rd;
               totalCountBytes += rd;

               if (rd != toread)
                  {
                     if (size - countBytes > 0)
                        {
                           sprintf(tmp, "At %d MiBytes: Short read error at offset %d, remaining %lld...\n",position,didRead,size-countBytes);
                           fprintf(flog, "%s", tmp);
                           if (fsummary)
                              fprintf(fsummary, "%s", tmp);
                           slog += tmp;

                           countUnreadable++;
                        }
                     else
                        {
                           fprintf(flog,"At %d MiBytes: End of file: short read error at offset %d.\n",position,didRead);
                           eof = true;
                        }
                  }
          
               if (readtime > gExtremelySlowReadTime)
                  {
                     sprintf(tmp, "At %d MiBytes: Extremely slow read at offset %d: %.3f sec, giving up.\n",position,didRead,readtime);
                     fprintf(flog, "%s", tmp);
                     if (fsummary)
                        fprintf(fsummary, "%s", tmp);
                     slog += tmp;

                     countSlow++;
                     countUnreadable++;
                     countAborted++;
                     eof = true;
                     break;
                  }

               countbytes += rd;
               counttime  += readtime;

#if 0
               printf("%d %f\n",rd,readtime);
               if (i>100)
                  {
                     stop = 1;
                     break;
                  }
#endif
               //usleep(1000);
            }

         double tend = dtime();

         if (haveFadvise64)
            {
               posix_fadvise64(rfd,seek,didRead,POSIX_FADV_DONTNEED);
            }

         //double rate = countbytes/counttime;
         double rate = countbytes/(tend-tstart);

         if (rate > maxRate)
            maxRate = rate;
               
         if (rate < minRate)
            minRate = rate;

         xcountbytes += countbytes;
         xcounttime += (tend-tstart);

         if (1)
            {
               static double xreadtime = 0;
               static time_t next_report = 0;
               time_t now = time(NULL);

               static int    xreads = 0;
               static double xbytes = 0;
               static double xtime = 0;

               if (countbytes > 0)
                  {
                     if (xreadtime < maxreadtime)
                        xreadtime = maxreadtime;

                     xreads += 1;
                     xbytes += countbytes;
                     xtime  += (tend - tstart);
                  }

               bool report = false;

               if (now >= next_report)
                  report = true;

               if (report && xbytes > 0)
                  {
                     if (next_report < now)
                        next_report = now;

                     next_report += 1;

                     double xrate = xbytes/xtime;

                     if (verbose)
                        fprintf(flog, "At %d MiBytes: %d reads, %.0f bytes, max read times: %.3f and %.3f sec, data rate: %.0f MiBytes/sec\r", position, xreads, xbytes, xreadtime, longestReadTime, xrate/mib);

                     xreadtime = 0;
                     xbytes = 0;
                     xtime = 0;
                     xreads = 0;
                  }
            }

         if (arg_time > 0)
            {
               double now = dtime();
               double timeTarget = timeBegin + arg_time*(double)((double)position*mib-(double)arg_offset)/(arg_size-(double)arg_offset);
               double diff = timeTarget - now;

               //printf("time begin %d, time end %d, target time %.1f, now %.1f, diff %.1f\n", timeBegin, timeTargetEnd, timeTarget, now, diff);

               if (diff > 0.1)
                  dsleep(diff);
            }

         position += incr;

         if (position*mib > arg_size)
            break;
      }

   time_t timeEnd;
   time(&timeEnd);

   close(rfd);

   double aveRate = xcountbytes/xcounttime;
   double effRate = (totalCountBytes/(double)(timeEnd-timeBegin));

   xctime(&timeEnd, nowText, sizeof(nowText));

   char summaryText[100*1024];

   sprintf(summaryText, "Disk \'%s\' size %.1f GiBytes, %d slow reads, %d unreadable areas, %d aborted, data rate: min: %.0f, max: %.0f, average: %.0f MiBytes/sec, effective: %.0f MiBytes/sec, elapsed %d sec, longest read time %.3f sec", diskname, size/(double)gib, countSlow, countUnreadable, countAborted, minRate/mib, maxRate/mib, aveRate/mib, effRate/mib, (int)(timeEnd-timeBegin), longestReadTime);

   fprintf(flog, "%s: %s\n", nowText, summaryText);

   if (fsummary)
      fprintf(fsummary, "%s: %s\n", nowText, summaryText);

   if (fxml)
      {
         char hostname[100*1024];
         gethostname(hostname, sizeof(hostname));

         std::string xml1 = "";
         xml1 += "<disk>\n";
         xml1 += xmlEncodeString("host", hostname) + "\n";
         xml1 += xmlEncodeString("name", diskname) + "\n";
         xml1 += xmlEncodeUint64("size", size) + "\n";
         xml1 += xmlEncodeInt("timeBegin", timeBegin) + "\n";
         xml1 += xmlEncodeInt("countSlow", countSlow) + "\n";
         xml1 += xmlEncodeInt("countUnreadable", countUnreadable) + "\n";
         xml1 += xmlEncodeInt("countAborted", countAborted) + "\n";
         xml1 += xmlEncodeDouble("minRate", minRate) + "\n";
         xml1 += xmlEncodeDouble("maxRate", maxRate) + "\n";
         xml1 += xmlEncodeDouble("aveRate", aveRate) + "\n";
         xml1 += xmlEncodeDouble("effRate", effRate) + "\n";
         xml1 += xmlEncodeDouble("longestReadTime", longestReadTime) + "\n";
         xml1 += xmlEncodeInt("timeEnd", timeEnd) + "\n";
         xml1 += "</disk>\n";
         
         fprintf(fxml, "%s", xml1.c_str());
      }

   bool doSendmail = gDoEmailAlways;

   if (countSlow && gDoEmailUnhappy)
      doSendmail = true;

   if ((countAborted||countUnreadable) && gDoEmailError)
      doSendmail = true;

   if (doSendmail && gEmailAddress.length() > 0)
      {
         char emailCmd[100*1024];
         char emailSubject[100*1024];
         char hostname[100*1024];

         gethostname(hostname, sizeof(hostname));

         sprintf(emailSubject, "diskscrub %s:%s - %d unreadable, %d slow", hostname, diskname, countUnreadable, countSlow);

         if (countAborted)
            strcat(emailSubject, ", aborted");

         sprintf(emailCmd, "mail -s \'%s\' %s", emailSubject, gEmailAddress.c_str());

         FILE *femail = popen(emailCmd, "w");
         fprintf(femail, "Date: %s\n", nowText);
         fprintf(femail, "Host: %s\n", hostname);
         fprintf(femail, "Disk: %s\n", diskname);
         fprintf(femail, "Size: %.1f GiBytes\n",  size/(double)gib);
         fprintf(femail, "Scan: %.1f GiBytes in %d sec (%.1f hrs) at %.0f MiBytes/sec\n", totalCountBytes/(double)gib, (int)(timeEnd-timeBegin), (double)(timeEnd-timeBegin)/3600, effRate/mib);
         fprintf(femail, "\n");
         fprintf(femail, "Aborted:    %d\n", countAborted);
         fprintf(femail, "Unreadable: %d\n", countUnreadable);
         fprintf(femail, "Slow:       %d\n",  countSlow);
         fprintf(femail, "Slowest read time: %.3f sec\n", longestReadTime);
         fprintf(femail, "\n");
         fprintf(femail, "Data rates: min: %.0f, max: %.0f, average: %.0f MiBytes/sec\n", minRate/mib, maxRate/mib, aveRate/mib);

         if (slog.length() > 0) {
            fprintf(femail, "\n");
            fprintf(femail, "Error log:\n");
            fprintf(femail, "%s", slog.c_str());
         }
         fprintf(femail, "\n");
         fprintf(femail, "--\n");
         fprintf(femail, "diskscrub\n");
         pclose(femail);
      }

   return 0;
}

struct disk_info {
   std::string name;
   uint64_t offset;
   uint64_t size;
   int retvalue;
};

struct disk_list {
   std::vector<disk_info> disks;

   void add(FILE* flog, const char* name, int sizeMiB=0, int offsetMiB=0)
   {
      disk_info disk;
      disk.name = name;
      disk.offset = offsetMiB*mib;
      disk.size   = sizeMiB*mib;

      int rfd = open_disk(name, flog);
      if (rfd < 0)
         return;

      if (disk.size == 0)
         disk.size = fsize(rfd);

      close(rfd);

      fprintf(flog, "Added \'%s\' size %lld, offset %lld\n", disk.name.c_str(), disk.size, disk.offset);

      disks.push_back(disk);
   }
};

void usage(const char* const argv[])
{
   fprintf(stderr,"Usage: %s [options...] /dev/sdXXX [offsetMiB] [sizeMiB] [/dev/sdYYY [off] [size]] ...\n",argv[0]);

   fprintf(stderr,"Options:\n");
   fprintf(stderr,"  --scrub-time TTT   : time allocated for the scan, in seconds. Value 0 means scan as quickly as possible\n");
   fprintf(stderr,"  --scrub-time-days DDD : time allocated for the scan, in days\n");
   fprintf(stderr,"  --readsize-MiB NNN : change size of read requests (recommended to use default value)\n");
   fprintf(stderr,"  -o FILENAME     : open given log file (default is stdout)\n");
   fprintf(stderr,"  -a FILENAME     : append to given log file (default is stdout)\n");
   fprintf(stderr,"  --summary FILENAME : append scrub summary into FILENAME\n");
   fprintf(stderr,"  --xml FILENAME     : write scrub results into an XML file\n");
   fprintf(stderr,"  --ionice           : reduce the scrub i/o priority\n");
   fprintf(stderr,"  --email-to USER@ADDRESS : send email to this address (use \"--email-to root\" for local root)\n");
   fprintf(stderr,"  --email-always  : always send email with scrub report\n");
   fprintf(stderr,"  --email-error   : send email report if disk error, etc\n");
   fprintf(stderr,"  --email-unhappy : also send email report if unhappiness\n");
   fprintf(stderr,"  --no-email-error   : no send email if disk error, etc\n");
   fprintf(stderr,"  --no-email-unhappy : no send email report if unhappiness\n");
   fprintf(stderr,"  --progress      : report progress\n");
   fprintf(stderr,"  --too-many-ioerrors NNN : stop after too many consecutive I/O errors (default is %d)\n", gTooManyIOerrors);
   fprintf(stderr,"  --too-slow-read-time TTT : stop if a read takes too long (default is %.0f, in sec)\n", gExtremelySlowReadTime);
   fprintf(stderr,"  --slow-read-time TTT : complain about reads that take too long (default is %.0f, in sec)\n", gSlowReadTime);

   exit(1);
}

int main(int argc, char*argv[])
{
   int status;

   if (argc < 2)
      {
         usage(argv); // DOES NOT RETURN
      }
   
   setbuf(stdout, 0);
   setbuf(stderr, 0);
   
   bool doIonice = false;
   bool doProgress = false;
   
   FILE* flog = stdout;
   FILE* fsummary = NULL;
   FILE* fxml = NULL;

   std::string xmlFilename, xmlFilenameTmp;
   
   time_t timeSec = 1;
   time_t timeMin = 60*timeSec;
   time_t timeHrs = 60*timeMin;
   time_t timeDay = 24*timeHrs;
   
   time_t cycle_time = 0; //6*timeDay;

   int readsize = 10 * mib;
   
   disk_list disks;
   
   //printf("argc %d\n", argc);
   
   if (argc == 2 && argv[1][0] != '-') {
      cycle_time = 0;
      disks.add(flog, argv[1]);
   } else if (argc == 3 && argv[1][0] != '-') {
      cycle_time = 0;
      disks.add(flog, argv[1], atoi(argv[2]));
   } else if (argc == 4 && argv[1][0] != '-') {
      cycle_time = 0;
      disks.add(flog, argv[1], atoi(argv[2]), atoi(argv[3]));
   } else {

      for (int iarg=1; iarg < argc; iarg++)
         {
            const char* arg = argv[iarg];
            //printf("argv[%d] is \'%s\'\n", iarg, arg);

            if (strcmp(arg, "-h")==0) {
               usage(argv); // does not return
            } else if (strcmp(arg, "--help")==0) {
               usage(argv); // does not return
            } else if (strcmp(arg, "--scrub-time")==0) {
               iarg++;
               cycle_time = atoi(argv[iarg]);
            } else if (strcmp(arg, "--scrub-time-days")==0) {
               iarg++;
               cycle_time = atoi(argv[iarg]) * timeDay;
            } else if (strcmp(arg, "--readsize-MiB")==0) {
               iarg++;
               readsize = atoi(argv[iarg]) * mib;
            } else if (strcmp(arg, "--too-many-ioerrors")==0) {
               iarg++;
               gTooManyIOerrors = atoi(argv[iarg]);
            } else if (strcmp(arg, "--too-slow-read-time")==0) {
               iarg++;
               gExtremelySlowReadTime = atof(argv[iarg]);
            } else if (strcmp(arg, "--slow-read-time")==0) {
               iarg++;
               gSlowReadTime = atof(argv[iarg]);
            } else if (strcmp(arg, "--progress")==0) {
               doProgress = true;
            } else if (strcmp(arg, "--ionice")==0) {
               doIonice = true;
            } else if (strcmp(arg, "-o")==0) {
               iarg++;
               const char* fname = argv[iarg];
               std::string x = fname;
               x += ".old";
               rename(fname, x.c_str());
               flog = fopen(fname, "w");
               if (flog == NULL) {
                  fprintf(stderr, "Cannot open log file \'%s\', errno %d (%s)\n", fname, errno, strerror(errno));
                  exit(1);
               }
               setbuf(flog, NULL);
            } else if (strcmp(arg, "-a")==0) {
               iarg++;
               const char* fname = argv[iarg];
               flog = fopen(fname, "a");
               if (flog == NULL) {
                  fprintf(stderr, "Cannot open log file \'%s\', errno %d (%s)\n", fname, errno, strerror(errno));
                  exit(1);
               }
               setbuf(flog, NULL);
            } else if (strcmp(arg, "--summary")==0) {
               iarg++;
               const char* fname = argv[iarg];
               fsummary = fopen(fname, "a");
               if (flog == NULL) {
                  fprintf(stderr, "Cannot open summary file \'%s\', errno %d (%s)\n", fname, errno, strerror(errno));
                  exit(1);
               }
               setbuf(fsummary, NULL);
            } else if (strcmp(arg, "--xml")==0) {
               iarg++;
               xmlFilename = argv[iarg];
               xmlFilenameTmp = xmlFilename + ".tmp";
               fxml = fopen(xmlFilenameTmp.c_str(), "w");
               if (fxml == NULL) {
                  fprintf(stderr, "Cannot open xml file \'%s\', errno %d (%s)\n", xmlFilenameTmp.c_str(), errno, strerror(errno));
                  exit(1);
               }
               setbuf(fxml, NULL);
               fprintf(fxml, "<scrub>\n");
            } else if (strcmp(arg, "--email-always")==0) {
               gDoEmailAlways  = true;
            } else if (strcmp(arg, "--email-unhappy")==0) {
               gDoEmailUnhappy = true;
               gDoEmailError = true;
            } else if (strcmp(arg, "--email-error")==0) {
               gDoEmailError = true;
            } else if (strcmp(arg, "--no-email-unhappy")==0) {
               gDoEmailUnhappy = false;
            } else if (strcmp(arg, "--no-email-error")==0) {
               gDoEmailError = false;
            } else if (strcmp(arg, "--email-to")==0) {
               iarg++;
               gEmailAddress = argv[iarg];
            } else if (arg[0] != '-') {
               const char* diskname = arg;

               int size = 0;
               int offset = 0;

               if (iarg+1<argc && isdigit(argv[iarg+1][0])) {
                  iarg++;
                  size = atoi(argv[iarg]);
               }
               
               if (iarg+1<argc && isdigit(argv[iarg+1][0])) {
                  iarg++;
                  offset = atoi(argv[iarg]);
               }
               
               disks.add(flog, diskname, size, offset);

            } else {
               usage(argv); // DOES NOT RETURN
            }
         }
   }

#ifdef HAVE_IOPRIO
   if (doIonice) {
      status = ioprio_set(IOPRIO_WHO_PROCESS, 0, IOPRIO_PRIO_VALUE(IOPRIO_CLASS_IDLE, 0));
      //status = ioprio_set(IOPRIO_WHO_PROCESS, 0, IOPRIO_PRIO_VALUE(IOPRIO_CLASS_BE, 7));
      if (status != 0) {
         fprintf(stderr, "Cannot ionice: ioprio_set() status %d, errno %d (%s)\n", status, errno, strerror(errno));
         exit(1);
      }
   }
#endif
   
   int tty = isatty(fileno(flog));

   if (tty)
      doProgress = true;
   
   uint64_t total_size = 0;
   for (unsigned int i=0; i<disks.disks.size(); i++)
      total_size += disks.disks[i].size;
   
   double rate = (double)total_size/(double)cycle_time;
   
   char nowText[1024];
   time_t now = time(NULL);
   xctime(&now, nowText, sizeof(nowText));

   char logText[100*1024];

   sprintf(logText, "Total size %lld, read size %d, cycle time %d sec, data rate %.1f MiBytes/sec", total_size, readsize, (int)cycle_time, rate/mib);

   fprintf(flog, "%s: %s\n", nowText, logText);

   if (fsummary)
      fprintf(fsummary, "%s: %s\n", nowText, logText);
   
   for (unsigned int i=0; i<disks.disks.size(); i++) {
      
      double portion = (double)disks.disks[i].size/(double)total_size;
      time_t t = (time_t)((double)cycle_time*portion);
      
      int retval = scrub_disk(flog, fsummary, fxml, doProgress, disks.disks[i].name.c_str(), disks.disks[i].offset, disks.disks[i].size, readsize, t);
      
      disks.disks[i].retvalue = retval;
   }

   if (fsummary)
      fclose(fsummary);

   if (fxml) {
      fprintf(fxml, "</scrub>\n");
      fclose(fxml);
      
      status = rename(xmlFilenameTmp.c_str(), xmlFilename.c_str());
      if (status != 0) {
         fprintf(flog, "Cannot rename xml file \'%s\' to \'%s\', errno %d (%s)\n", xmlFilenameTmp.c_str(), xmlFilename.c_str(), errno, strerror(errno));
      }
   }

   fclose(flog);

   return 0;
}

// end file
